/*
** Module   : text.g
** Abstract : This grammar is the top level recognizer of the Progress
**            language preprocessor input. It recognizes tabs, nested
**            comments, single and double quoted multiline strings and
**            the code.
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20041112   @18832 Created. This grammar is the top level reco-
**                           gnizer of the Progress language preprocessor
**                           input. It recognizes tabs, nested comments,
**                           single and double quoted multiline strings
**                           and the code. It recognizes the preprocessor
**                           statements far enough to be able to switch
**                           to a specialized statements parser.
** 002 NVS 20041222   @19050 For some files, the parser stops just before 
**                           the final NL and no final NL appears on the 
**                           output. Modified the "text" rule so now it 
**                           matches the final NL and EOF. This modifica-
**                           tion caused creation of another rule 
**                           textBlock which is the equivalent of the old 
**                           text rule, but now features greedy option for
**                           the loop closure.
** 003 NVS 20041222   @19051 ASTMT and CODE rules were designed to find 
**                           &text anywhere in the input stream. This 
**                           breaks the Progress symbols that may contain 
**                           '&' inside. Redesigned the CODE rule to not 
**                           detect internal ampersands in text. Only 
**                           white space separated words starting with '&'
**                           are considered preprocessor statements now. 
** 004 NVS 20041222   @19031 The toplevel.test testcase shows that nested
**                           &IFs weare processed incorrectly. A block to 
**                           be skipped causes a problem if it contains 
**                           nested &IF. The skipBlock rule stops at the 
**                           closest &ELSEIF or &ELSE or &ENDIF even if it
**                           belongs to the nested &IF. Fix: Added new 
**                           skipIf rule that matches noninterpretable 
**                           &IF...&ENDIF sequence, possibly recursive. 
**                           Changed skipBlock to include skipIf as one of
**                           the two alternatives. 
** 005 NVS 20041222   @19061 inString signal from the lexer to the 
**                           ClearStream is always late dute to the look-
**                           ahead (k=2) in the lexer. Conversion of new-
**                           lines inside strings is reworked to use 
**                           escape characters inserted into the stream. 
**                           CR is used to mark all converted CRs and NLs.
**                           Original CRs are always removed from the 
**                           stream. Original NLs are always removed from 
**                           strings. Late signalling makes it possible 
**                           for the escapes that immediately follow the 
**                           string like in "..."~n to produce "..." CR NL
**                           although this normally shouldn't happen. The 
**                           lexer checks to see if the lookahead buffer 
**                           LA(1) is CR after matching the closing chara-
**                           cter of the string and consumes it fixing the
**                           problem.
** 006 NVS 20050127   @19415 Lexer vocabulary was for 0x03-0xff which was
**                           causing problems converting "~000". Fixed to
**                           be the full range 0x00-0xff.
** 007 NVS 20050127   @19418 ASTRING and QSTRING rules are fixed to handle
**                           the cases with double tildes like "~~". Pre-
**                           iously they just ignored the first tilde as a
**                           stand alone character and treated the second
**                           tilde as an escape ~" which led into an exce-
**                           ption as the string were treated as unclosed.
**                           Now they properly identify double tildes and
**                           consume them.
** 008 NVS 20050209   @19736 Fixed tab expansion bug. The setColumn() call
**                           in ClearStream is always ahead of what text.g
**                           sees as the current column due to lookahead
**                           processing. The tab() method saw a wrong
**                           column and calculated wrong number of spaces
**                           to insert. Adding newline() calls back to
**                           WS, COMMENT, ASTRING and QSTRING fixes this
**                           deviation.
** 009 GES 20050502   @20991 Added support for '\\' as an escape char
**                           (equivalent to tilde in all ways) inside
**                           ASTRING and QSTRING.
** 010 NVS 20050513   @21165 Many methods now throw IOException. Added to
**                           method headers and javadoc. Added calls to
**                           setInStatement method to control markers
**                           processing while parsing preprocessor
**                           statements.
** 011 GES 20060904   @29213 Added processing to space-ify string literals
**                           with an embedded null character escape
**                           sequence.
** 012 GES 20070319   @32448 Fix for IF/ELSE where ELSE would always emit
**                           even when the IF portion already emitted.
** 013 GES 20070403   @32713 Minor signature change.
** 014 GES 20070923   @35188 Print a stack trace on errors.
** 015 GES 20070926   @35240 Modified string parsing to ignore backslash
**                           as an escape character if unix escapes are
**                           not being honored.
** 016 GES 20071204   @36197 Turn off braces expansion within a skipped
**                           conditional block. This allows include file
**                           recursion and avoids possible bad behavior.
**                           Refactored lookahead and matching for
**                           comments to enable proper braces processing
**                           in between the open '/' and '*' chars. This
**                           was used to enable/disable comments based on
**                           arguments to an include file but was broken
**                           since marker processing was not possible
**                           between these two characters.
** 017 GES 20080318   @37643 Changed approach to read characters (the
**                           entire unicode charset) instead of dealing
**                           with bytes. This makes it safe to process
**                           files with I18N requirements. Disable the
**                           in-statement flag when processing a 
**                           conditional block (only enable the flag
**                           while the preprocessor directive/expression
**                           is being handled).
** 018 GES 20081010   @40336 Fix to allow the ampersand char inside code.
** 019 GES 20110628          Combined text.g and stat.g into a single lexer
**                           and parser implementation. This eliminates the
**                           need to switch/multiplex lexers and it makes the
**                           preprocessor much simpler. It also solves problems
**                           with lost lookahead tokens in certain situations
**                           when switching lexers. Added broken mode for
**                           string processing in global or scoped defines.
**                           Added filtering of markers for global and
**                           scoped defines (the definition part) since it is
**                           possible to have nested includes there. Without
**                           the filtering, the marker characters are left as
**                           part of the definition.  When used, the define
**                           would trigger the failsafe code in ClearStream
**                           which aborts when reading a marker char. Made the
**                           ASTMT lexer rule only ever return a CODE token or
**                           one of the known preproc statements. This allows
**                           ampersand characters to be embedded in random text
**                           (like a variable name!) which is legal in 4GL.
** 020 GES 20111206          Fixed 3 issues: A COMMENT matched as CODE that
**                           is followed by more CODE (with no other token
**                           type like WS in between) would fail to return
**                           just the comment. An &UNDEFINE as the last text in
**                           an include file would try to undefine the name
**                           on the previous scope instead of the scope that
**                           was in effect when the UNDEFINE was read. Progress
**                           tolerates a missing &ENDIF if EOF is encountered
**                           before the &IF is closed (this is just undefined
**                           behavior and wasn't really a bug).
*/

header {
/*
** Module   : TextLexer.java
**            TextParser.java
**            PreprocTokenTypes.java
** Abstract : This grammar is the top level recognizer of the Progress
**            language preprocessor input. It recognizes tabs, nested
**            comments, single and double quoted multiline strings and
**            the code.
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20041116 ADD   @18832 WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is text.g!
*/

package com.goldencode.p2j.preproc;

import java.io.*;
import java.util.*;
import antlr.collections.AST;
import antlr.collections.impl.*;
import antlr.debug.misc.*;
import antlr.*;
import com.goldencode.p2j.util.*;
}

/**
 * Declares the entry point class of this grammar.
 * This grammar takes a Progress source file and starts parsing it
 * to a degree necessary for performing the preprocessing only.
 * The preprocessing is done on the go without creating an AST tree
 * first. This is possible due to a limited amount of transformations
 * that has to be done to the input.
 * <p>
 * The parsing starts when a driving code calls the
 * <code>{@link #text()}</code>
 * method of this class. The processed output is written to an output
 * stream as the parsing progresses. As soon as the parser reaches
 * the end of file, the preprocessor is done.
 * <p>
 * The parser calls the lexer's <code>nextToken</code> method and works
 * with the stream of <code>Token</code> objects. So, all the details
 * of tokenizations are hidden in the lexer. The lexer, in turn, reads
 * its input character by character from an input stream. The input
 * stream is an instance of the <code>{@link ClearStream}</code> class
 * that hides from the lexer some essential complexity of the preprocessing
 * like handling the alternative codings, escape sequences and substitutions
 * for braces.
 * <p>
 * The parser recognizes and properly handles the following constructs:
 * <ul>
 *   <li>regular Progress code on input is copied to output according
 *       to the conditional generation currently in effect;
 *   <li>tabs are expanded on output;
 *   <li>other white spaces are preserved on output;
 *   <li>Progress language strings and comments are recognized because
 *       they affect the workings of the preprocessor;
 *   <li>preprocessor statements are recognized and processed;
 * </ul>
 * @see  TextLexer
 * @see  BracesParser
 * @see  BracesLexer
 */
class TextParser extends Parser;

options
{
   k = 2;                                        // lookahead depth is 2
   importVocab = Shared;
   exportVocab = Preproc;
}

// parser class extensions
{
   /** Keeps a reference to the shared environment. */
   private Environment env = null;
   
   /** Nesting level for text block processing. */
   private int nested = 0;
   
   /** Tracks nesting level for conditional preproc directives. */
   private int condNesting = 0;

   /**
    * Constructor. Creates a parser attached to the token stream
    * selector taken from the environment. Saves the environment
    * for future needs.
    *
    * @param    lexer
    *           The lexer for this instance.
    * @param    env
    *           The shared preprocessor environment.
    */
   public TextParser(TokenStream lexer, Environment env)
   {
      this(lexer);
      this.env = env;
      this.env.setPsi(this.getInputState());
   }
   
   /**
    * Writes error data to <code>System.err</code>, including a full stack
    * trace.
    *
    * @param    re
    *           The error on which to report.
    */
   public void reportError(RecognitionException re)
   {
      re.printStackTrace(System.err);
   }
}

/**
 * Matches any interpretable block of text from the Progress source
 * file. This can be either the top level source text or any block
 * of text following <code>&THEN</code> or <code>&ELSE</code>
 * statements when their appropriate preprocessor expressions evaluate
 * to those clauses.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 * @throws      IOException
 */
text
throws IOException
   :
      textBlock
      EOF
   ;

/**
 * Matches any interpretable block of text from the Progress source
 * file. This can be either the top level source text or any block
 * of text following <code>&THEN</code> or <code>&ELSE</code>
 * statements when their appropriate preprocessor expressions evaluate
 * to those clauses.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 * @throws      IOException
 */
textBlock
throws IOException
   :
      {
         // track our nesting level
         nested++;
         
         // cache off the in stmt flag
         boolean inStmt = env.isInStatement();
         
         // are we nested? (this can only happen due to an &IF)
         if (nested > 1 && inStmt)
         {
            // disable while here
            env.setInStatement(false);
            env.clearDeferred();
         }
      }
      (
         options
         { 
            generateAmbigWarnings = false;
            greedy = true;
         }
         :
           // matches all ampersand prefaced preprocessor directives
           ppstatement
           
           // matches everything else
         | t:~(AELSEIF | AELSE | AENDIF)
           {
              env.print(t.getText());
           }
      )*
      {
         // if we are in a nested block
         if (nested > 1)
         {
            // restore our saved state
            env.setInStatement(inStmt);
         }
      
         // clear our nesting level
         nested--;
      }
   ;

/**
 * Matches any Progress preprocessor statement, even mistyped one.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 * @throws      IOException
 */
ppstatement
throws IOException
   :
      {
         if (!env.isInStatement())
         {
            env.setInStatement(true);
         }
      }
      (
           globdef
         | scopdef
         | undefine
         | message
         | noop
         | aif
         | astmt
      )
      {
         env.setInStatement(false);
         env.clearDeferred();
      }
   ;

/**
 * Matches the <code>&GLOBAL-DEFINE</code> Progress preprocessor statement.
 */
globdef
throws IOException
   :
      AGLOBAL define[true]
   ;

/**
 * Matches the <code>&SCOPED-DEFINE</code> Progress preprocessor statement.
 */
scopdef
throws IOException
   :
      ASCOPED define[false]
   ;

/**
 * Matches <code>&GLOBAL-DEFINE</code> and <code>&SCOPED-DEFINE</code>
 * preprocessor directives and creates newly defined preprocessor variable
 * in the corresponding scope. This consumes everything up to the end of
 * the line and uses the text (trimmed of spaces at the front and back) as
 * the value of the preprocessor variable.
 *
 * @param    global
 *           <code>true</code> if this is a <code>&GLOBAL-DEFINE</code> and
 *           <code>false</code> for <code>&SCOPED-DEFINE</code>.
 */
define [boolean global]
throws IOException
   {
      String name = null;
      StringBuilder definition = new StringBuilder();
      int line = 0;
      int column = 0;
   }
   :
      (WS)+
      {
         line   = env.getLex().getLine();
         column = env.getLex().getColumn();
      }
      name = ppname
      (
         options { greedy=false; generateAmbigWarnings=false; }
         :
           // remove comments from the definition, but emit them in the output
           c:COMMENT! { env.print(c.getText()); }
           
           // accumulate the definition
         | def:~(NL)  { definition.append(def.getText()); }
      )*
      NL
      
      // add the definition to the scoped symbol dictionary
      {
         // the definition may have pending markers, process them
         StringBuffer unclean = new StringBuffer(definition);
         definition = new StringBuilder(env.filter(unclean, false));
      
         // remove any leading blanks
         while (definition.length() > 0 && definition.charAt(0) == ' ')
            definition.deleteCharAt(0);
   
         // remove any trailing blanks
         while (definition.length() > 0 && 
               definition.charAt(definition.length() - 1) == ' ')
            definition.setLength(definition.length() - 1);
   
         // remove any tildes that may have been left in '~{' pairs
         // while(definition.length() > 0 && definition.indexOf("~{") >= 0)
         //    definition.deleteCharAt(definition.indexOf("~{"));

         // as far as we know, there are no real syntax rules for the name
         // so there is no known checks that need to be applied here
         
         // lookup the name to see if it already exists
         ScopedSymbolDictionary sym = env.getSym();
         Symbol cur = (Symbol) sym.lookupSymbol(name);

         // check to see if the name is for a builtin variable
         if (cur != null && cur.getOrigin() == Symbol.BUILTIN)
         {
            env.eprint("cannot redefine builtin variable " + name,
                       line,
                       column);
            return;
         }

         if (global)
         {
            // global definitions can be added if they don't already exist
            // and they can be overridden if they only exist as global defs
            // (but the override will fail if the name already exists as
            // a scoped definition or as an argument); builtins can't be 
            // overridden either but they are handled above
            if (cur == null || cur.getOrigin() == Symbol.GLOBAL)
            {
               sym.addSymbol(true,
                             name,
                             new Symbol(Symbol.GLOBAL, definition));
            }
            else
            {
               // TODO: Check if this "silent bypass if the values are the
               //       same" logic is actually how the 4GL preproc works.
               //       This behavior was found in a customer app and this is
               //       just a theory.
               if (!cur.getValue().equals(definition.toString()))
               {
                  String spec =
                     "Invalid global redefinition of preprocessor name '%s'." +
                     " (2967)";
                     
                  env.eprint(String.format(spec, name), line, column);
               }
            }
         }
         else
         {
            // put the symbol definition into the locally scoped dictionary
            sym.addSymbol(false,
                          name,
                          new Symbol(Symbol.SCOPED, definition));
         }
      }
   ;
   
/**
  * Matches <code>&UNDEFINE</code> preprocessor statement and removes
  * the first found definition of the preprocessor variable in the nearest
  * scope. Contrary to the parsing of {@link #define} which consumes all
  * text up to the end of the line, this directive only consumes whitespace
  * followed by a single preprocessor name. 
  */
undefine
throws IOException
   {
      String name   = null;
      int    line   = 0;
      int    column = 0;
      
      // after the ppname rule is called, the ClearStream may have already
      // popped our scope, so we cache the possible names here so that we can
      // check if the name was already removed because the scope was popped;
      // this occurs when the UNDEFINE directive is the very last text in
      // an include file, which causes the popping to occur before the name
      // is parsed here
      ScopedSymbolDictionary ssd = env.getSym();
      Set<String> poppedNames = ssd.keySet(0);
   }
   :
      AUNDEFINE
      (
         options { generateAmbigWarnings=false; }
         :      
           WS
         | c:COMMENT! { env.print(c.getText()); }
      )+
      {
         line   = env.getLex().getLine();
         column = env.getLex().getColumn();
      }
      name = ppname
      {
         String pkey = ssd.processKey(name);
      
         // undefine the nearest definition of the symbol, if not defined,
         // issue an error message
         if (!ssd.deleteSymbol(name) && !poppedNames.contains(pkey))
         {
            env.eprint("symbol '" + name + "' not defined.", line, column);
         }
      }
   ;

/**
  * Matches <code>&MESSAGE</code> preprocessor statement and prints
  * the message text into the error stream without incrementing the errors
  * count.
  */
message
throws IOException
   {
      StringBuffer rest = new StringBuffer();
   }
   :
      AMESSAGE
      (
         options { generateAmbigWarnings=false; }
         :
         WS | c:COMMENT! { env.print(c.getText()); }
      )+
      (
         options { greedy=false; }
         :
           /* comments are output to stream, but otherwise ignored */
           c2:COMMENT!    { env.print(c2.getText()); }
           
           // accumulate the rest of line
         | rst:~NL        { rest.append(rst.getText()); }
      )*
      nl:NL
      {
         env.mprint(rest, nl.getLine());
      }
   ;

/**
  * Matches the <code>&ANALYZE-SUSPEND</code> and <code>&ANALYZE-RESUME</code>
  * preprocessor statements.  Everything is parsed to the end of the line
  * and is dropped.  The newline is preserved in the output.
  */
noop
throws IOException
   :
      (ASUSPEND | ARESUME) 
      (
         options { greedy=false; }
         :
           /* comments are output to stream, but otherwise ignored */
           c:COMMENT!     { env.print(c.getText()); }
           
         | ~NL
      )*
      NL
      {
         env.print("\n");
      }
   ;

/**
  * Matches sequence of tokens that make a preprocessor variable name.
  *
  * @return  the gathered variable name as a string
  *
  * @throws      antlr.RecognitionException
  * @throws      antlr.TokenStreamException
  * @throws      IOException
  */
ppname returns [String s]
throws IOException
   {
      StringBuilder buf = new StringBuilder();
      s = null;
   }
   :
      (
         options { generateAmbigWarnings=false; }
         :
           /* comments are output to stream but are not included in the name */
           c:COMMENT!     { env.print(c.getText()); }
           
           /* first whitespace or newline ends the name */
         | t:~(WS | NL)   { buf.append(t.getText()); }
      )+
      {
         s = buf.toString(); 
      }
   ;
   
/**
 * Matches the <code>&IF</code> Progress preprocessor statement.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 * @throws      IOException
 */
aif
throws IOException
   :
      tmp:AIF
      {
         condNesting++;
         // String filename = ((FileScope)(env.getSym().getScope())).getFileName();
         // System.out.printf("&IF level %d, filename %s, line %d, col %d\n", condNesting, filename, tmp.getLine(), tmp.getColumn());
      }
      condtext
   ;

/**
 * Concludes the conditional input parsing for <code>&IF &ENDIF</code>
 * preprocessor statements. Due to the nested nature of conditional
 * statements, this method may be called recursively.
 * <p>
 * The preprocessor expression evaluation is done in two steps.
 * Firstly, the entire expression is gathered from multiple lexer tokens
 * into a string by calling the <code>expression</code> method.
 * Secondly, if the expression has to be evaluated, the
 * <code>{@link Preprocessor#evaluate(Environment,String)}</code> method
 * is called to do the job.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 * @throws      IOException
 */
condtext
throws IOException
   // The following defines the code emission control variable meaning:
   // 0 - previous expression evaluated to FALSE or never evaluated;
   //     do not emit the code; evaluate the nearest expression
   // 1 - previous expression evaluated to TRUE;
   //     emit the code in the nearest block
   // 2 - previous expression evaluated to TRUE and the code block has been
   //      emitted; skip till &ENDIF
   // The value is selected so that emit variable either stays at 0
   // or changes once to 1 and 2 and stays at 2.
   {
      int emit = 0;
      String x = null;
   }
   :
      // gather the expression into a string
      x = expression[emit]
      {
         // evaluate the expression
         if (Preprocessor.evaluate(env, x))
         {
            // force the following block to emit
            emit = 1;
         }
         else
         {
            // disable include expansion; must be done here BEFORE any
            // matching may inadvertantly access lookahead and trigger
            // an expansion; this is OK as long as it is above the &THEN
            env.setExpandBraces(false);
         }
      }

      ATHEN
      {
         // System.out.printf("&THEN1\n");
      }                      

      // process the following block
      (
           { emit == 1 }?
           textBlock               // emit it
         | skipBlock               // drop it
      )

      (
         options { generateAmbigWarnings=false; }
         :
         tmp1:AELSEIF
         {
            // String filename = ((FileScope)(env.getSym().getScope())).getFileName();
            // System.out.printf("&ELSEIF level %d, filename %s, line %d, col %d\n", condNesting, filename, tmp1.getLine(), tmp1.getColumn());
         }
         
         // if the 'emit' var is 1 then it means that a previous block has
         // already been emitted and we are done, force this by setting
         // emit to 2
         {
            if (emit == 1)
               emit = 2;
         }
         
         // gather next expression into a string
         x = expression[emit]
         {
            // did the IF portion emit?
            if (emit == 0)
            {
               // no emit yet, so evaluate the expression
               if (Preprocessor.evaluate(env, x))
               {
                  // force the following block to emit
                  emit = 1;
               }
               else
               {
                  // disable include expansion; must be done here BEFORE any
                  // matching may inadvertantly access lookahead and trigger
                  // an expansion; this is OK as long as it is above the
                  // &THEN
                  env.setExpandBraces(false);
               }
            }
         }

         ATHEN
         {
            // System.out.printf("&THEN 2\n");
         }

         // process the following block
         (
              { emit == 1 }?
              textBlock             // emit
            | skipBlock             // drop
         )
      )*

      (
         options { generateAmbigWarnings=false; }
         :
         {
            // will we be dropping the block?
            if (emit != 0)
            {
               // disable include expansion; must be done here BEFORE any
               // matching may inadvertantly access lookahead and trigger
               // an expansion; this is OK as long as it is above the
               // &ELSE
               env.setExpandBraces(false);
            }
         }
      
         tmp2:AELSE
         {
            // String filename = ((FileScope)(env.getSym().getScope())).getFileName();
            // System.out.printf("&ELSE level %d, filename %s, line %d, col %d\n", condNesting, filename, tmp2.getLine(), tmp2.getColumn());
         }

         // process the following block, if emit is 0, then no prior block
         // has emitted, so we must emit, otherwise we drop the block
         (
              { emit == 0 }?
              textBlock             // emit
            | skipBlock             // drop
         )
      )?

      (
         options { generateAmbigWarnings=false; }
         :
           tmp3:AENDIF
         | { LA(1) == EOF }?
           // empty alternative (make &ENDIF optional) on EOF
      )
      {
         // String filename = ((FileScope)(env.getSym().getScope())).getFileName();
         // System.out.printf("&ENDIF level %d, filename %s, line %d, col %d\n", condNesting, filename, tmp3.getLine(), tmp3.getColumn());
         condNesting--;
      }
      // AENDIF                        // terminate this &IF
   ;

/**
 * Gathers the preprocessor expression into a string
 * or ignores it.
 *
 * @param   emit
 *          if 0, instructs the method to gather expression tokens
 *          into a string, otherwise the expression is ignored.
 * @return  the whole expression as a string if <code>emit</code>
 *          is 0, or "".
 */
expression[int emit] returns [String x]
   {
      StringBuilder buf = new StringBuilder();
      x = null;
   }
   :
      (
         t:~ATHEN
         {
            if (emit == 0)
            {
               buf.append(t.getText());
            }
         }
      )+
      {
         x = buf.toString();
         
         // if (emit == 0)
         //   System.out.printf("EXPRESSION: %s\n", x);
      }
   ;

/**
 * Matches non-interpretable block of text from the Progress source
 * file which is a fully coded <code>&IF &ENDIF</code> statement, possibly 
 * recursive. Non-interpretable blocks are the result of the preprocessor
 * expressions evaluated so they cause bypassing of corresponding
 * clauses of <code>&IF...&ENDIF</code> statements.
 * <p>
 * This is a simple replacement for the {@link #aif} rule which discards
 * any content and uses recursion to drop any nested conditional blocks.
 * See {@link #skipBlock} which is the only user of this rule.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 */
skipIf
   :
      tmp:AIF
      {
         condNesting++;
         // String filename = ((FileScope)(env.getSym().getScope())).getFileName();
         // System.out.printf("&IF (SKIPPING) level %d, filename %s, line %d, col %d\n", condNesting, filename, tmp.getLine(), tmp.getColumn());
      }
      (
         options { generateAmbigWarnings=false; }
         :
           skipIf
         | ~(AENDIF | EOF)
      )*
      (
         options { generateAmbigWarnings=false; }
         :
           tmp2:AENDIF
         | { LA(1) == EOF }?
           // empty alternative (make &ENDIF optional) on EOF
      )
      {
         // filename = ((FileScope)(env.getSym().getScope())).getFileName();
         // System.out.printf("&ENDIF (SKIPPING) level %d, filename %s, line %d, col %d\n", condNesting, filename, tmp2.getLine(), tmp2.getColumn());
         condNesting--;
      }
      
   ;
    
/**
 * Matches any non-interpretable block of text from the Progress source
 * file, including those containing nested <code>&IF &ENDIF</code> 
 * statements. Non-interpretable blocks are the result of the preprocessor
 * expressions evaluated so they cause bypassing of corresponding
 * clauses of <code>&IF...&ENDIF</code> statements.
 * <p>
 * It is not possible to take the simple approach of matching everything 
 * until the next <code>&ELSEIF, &ELSE or &ENDIF</code> because there may be
 * a nested <code>&IF</code> inside this block. In such a case, the
 * <code>&IF</code> block will definitely have its own <code>&ENDIF</code>
 * (and may also have its own <code>&ELSEIF or &ELSE</code>) which should
 * not cause the termination of this loop since the entire 
 * <code>&IF...&ENDIF</code> is dead code. The matching loop can only
 * terminate when it reaches an <code>&ELSEIF, &ELSE or &ENDIF</code> that
 * is at the same level (no nesting) as the current block being skipped. The
 * {@link #skipIf} rule handles matching any nested <code>&IF</code>
 * constructs and uses recursion to handle any nested blocks that are
 * nested inside of the nested block and so on.
 * <p>
 * Note that braces expansion must be disabled before entering this rule
 * since the lookahead processing may trigger expansion of includes or
 * argument references before this rule can be executed. Since braces are
 * not expanded inside skipped conditional blocks in Progress, this is
 * valid. This behavior also enables include files to include themselves
 * recursively (so long as the <code>&IF</code>) blocks are VERY carefully
 * crafted.  This rule will always re-enable braces expansions before it
 * returns.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 */
skipBlock
   :
      (
         options { generateAmbigWarnings=false; }
         :
           skipIf
         | ~(AIF | AELSEIF | AELSE | AENDIF | EOF)
      )*
      {
         // re-enable include expansions
         env.setExpandBraces(true);
      }
   ;

/**                                                   
 * Matches any mistyped or unrecognized Progress preprocessor statement.
 *
 * @throws      antlr.RecognitionException
 * @throws      antlr.TokenStreamException
 */
astmt
   :
      s:ASTMT
      {
         env.eprint("unrecognized statement '" + s.getText() + "'",
                    s.getLine(), s.getColumn());
      }
   ;

/**
 * Tokenizes the input stream of characters from the Progress source
 * file and returns tokens to the caller according to the needs of
 * preprocessor.
 * <p>
 * The lexer recognizes the current context such as comments, strings
 * or code and keeps track of the context switches using
 * <code>{@link Environment#setInComment(boolean)}</code> and
 * <code>{@link Environment#setInString(boolean)}</code> methods.
 *
 * @see  TextParser
 * @see  Environment
 */
class TextLexer extends Lexer;

options
{
   k = 2;                                 // lookahead depth is 2
   charVocabulary = '\u0000'..'\ufffe';   // specifies range 0x0000-0xFFFE
                                          // (0xFFFF is EOF)
   importVocab = Shared;
   exportVocab = Preproc;
   testLiterals = true;
   caseSensitive = false;
   caseSensitiveLiterals = false;
   // codeGenBitsetTestThreshold=999; 
}

tokens
{
   AGLOBAL   = "&global-define";
   ASCOPED   = "&scoped-define";
   AMESSAGE  = "&message";
   AUNDEFINE = "&undefine";
   AIF       = "&if";
   ATHEN     = "&then";
   AELSEIF   = "&elseif";
   AELSE     = "&else";
   AENDIF    = "&endif";
   ASUSPEND  = "&analyze-suspend";
   ARESUME   = "&analyze-resume";
}

{
   /** keeps the reference to the shared environment */
   private Environment env = null;

   /** Nesting level for comments. */
   private int commentNesting = 0;
   
   /** Control matched quote processing in strings. */
   private boolean brokenStringMatching = false;
   
   /**
    * Constructor. Creates a lexer attached to the input stream
    * taken from the environment. Saves the environment
    * for future needs.
    *
    * @param    env
    *           Shared preprocessor environment.
    */
   public TextLexer(Environment env)
   {
      this(new CharBuffer(env.getIns()));
      this.env = env;
      this.env.setLsi(this.getInputState());
   }

   /**
     * Expands tabs with spaces right in the ANTLRStringBuilder text.
     * This version is used here because according to experiments,
     * the tabs expansion for the preprocessor variables is immediate.
     */
   public void tab()
   {
      int c  = getColumn();
      int nc = (((c-1)/tabsize) + 1) * tabsize + 1;  // calculate tab stop
      setColumn(nc);

      text.setLength(text.length() - 1);    // replace '\t' in the buffer
      for (; c < nc; c ++)                  // add more spaces
         text.append(' ');
   }

   /**
    * Tests the token text against the literals table and provides
    * correct token types for the abbreviated preprocessor statement
    * keywords.
    */
   public int testLiteralsTable(int ttype)
   {
      // support abbreviations in some of the possible literals
      String[] abbrs =
      {
         "&global-define",
         "&scoped-define",
         "&undefine"
      };
      
      // minimum lengths of each of the above possible abbreviations
      int[] minsz = {5, 5, 6};

      String  itoken = new String(text.getBuffer(), 0, text.length());
      String  token  = itoken.toLowerCase();
      int     ltoken = token.length();

      for (int i = 0; i < abbrs.length; i ++)
      {
         if (abbrs[i].startsWith(token) && ltoken >= minsz[i])
         {
            token = abbrs[i];
         }
      }

      hashString.setString(token);
      
      Integer literalsIndex = (Integer)literals.get(hashString);
      
      if (literalsIndex != null)
      {
         ttype = literalsIndex.intValue();
      }
      
      // disable normal quote matching in STRING until the NL is reached
      // (this is necessary since unmatched quotes are allowed in defines
      // and all downstream parsing is really broken otherwise)
      if (ttype == AGLOBAL || ttype == ASCOPED)
      {
         brokenStringMatching = true;
      }
      
      return ttype;
   }
}

/**
 * Matches any amount of continguous whitespace (spaces and tabs) in a program.
 * Newlines are NOT matched.
 * <p>
 * This is a  top level lexer rule which means that there is an associated
 * <code>WS</code> token.
 */
WS
   :
      (
         options { greedy = true; }
         :
           ' '
         | '\t'
      )+
   ;

/**
 * Matches a single newline character.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>NL</code> token.
 */
NL
   :
      '\n'
      {
         newline();
         
         // if a scoped or global define has ended, re-enable STRING processing
         if (brokenStringMatching)
         {
            brokenStringMatching = false;
         }
      }
   ;
   
/**
 * Matches Progress language strings. Brackets the string with a pair
 * of calls to <code>{@link Environment#setInString(boolean)}</code>
 * so the preprocessor always knows the context.
 * <p>
 * Depending on -keeptildes preprocessor options, newlines may be required
 * to be deleted from the strings. This requirement does not extend to the
 * characters which are coded as escape sequences, though. The final action in
 * this rule postprocesses the string contents so they comply with the 
 * requirement.
 * <p>
 * To tell the original newlines from the escaped ones, ClearStream uses the
 * following technique. For every converted character, that becomes newline,
 * an escape character is inserted in front of it into the stream. The 
 * postprocessing action is supposed to detect this escape character, discard
 * it but leave the newline that follows untouched. The original newlines come
 * without escape characters and are deleted.
 * <p>
 * For this technique to work, an escape character has to be assigned which
 * normally can't be seen in the stream. Such a character is CR, due to the
 * fact that ClearStream strips off all original CRs and replaces them with
 * NLs. But the same problem of converted CRs still exists. Fortunately, it is
 * solved in a similar way. As the result, the escape character, CR, is
 * inserted in front of every converted CR or NL. The valid escape sequences
 * that ClearStream produces are CR CR and CR NL. In both cases, the first CR
 * has to be removed from the stream and the second character left untouched.
 * <p>
 * One more complication is due to the fact that there is no on time state
 * change signalling between the lexer and the ClearStream code. Although this
 * rule signals the end of string by calling <code>inString(false)</code>, the
 * call is one character late because of the k=2 lookahead in the lexer. Late
 * signalling leaves a hole in the algorithm for a wrong escape character
 * insertion just outside the string, which should not happen. Consider the
 * following stream:
 * <pre>
 *    "..."~n~n
 * </pre>
 * Due to the late signalling, ClearStream reads and converts the ~n Progress
 * escape sequence before this rule signals the end of string. As the result,
 * ClearStream produces the following stream:
 * <pre>
 *    "..." CR NL NL
 * </pre>
 * Notice that only the first NL gets escaped with CR, the second NL comes
 * when the end of string signal has been received. This issue is fixed by
 * checking the next character in the lookahead buffer LA(1) in the
 * postprocessing rule. If it is CR, it has to be discarded, as it only can be
 * there as the result of this late signalling.
 * <p> 
 * This is a top level lexer rule which means that there is an associated
 * <code>STRING</code> token.
 * <p>
 * If there is an escaped <code>null</code> character in the string literal,
 * it must be "space-ified" by converting it and all following characters
 * into spaces (including the proper reduction of other following escape
 * sequences into a single space character).  One can encode a 
 * <code>null</code> escape sequence in a string literal in Progress BUT a
 * <code>null</code> character will NEVER end up in the result!  This method
 * duplicates that processing.
 *
 * @throws   antlr.RecognitionException
 * @throws   antlr.CharStreamException
 * @throws   antlr.TokenStreamException
 */
STRING
   :
      {
         if (!brokenStringMatching)
         {
            env.setInString(true);
         }
      }
      (ASTRING | QSTRING)
      {
         if (!brokenStringMatching)
         {
            env.setInString(false);
            
            text.setLength(Preprocessor.stripNL(text.getBuffer(),
                                                text.length(),
                                                env.getOpt().isKeepTildes()));
            if (LA(1) == 0x0D)
            {
               consume();
               text.setLength(text.length() - 1);
            }
            
            boolean unixEsc = env.getOpt().isUnixEscapes();
            
            // implement spacifying transformation here
            setText(character.progressSpacifyNull(getText(), unixEsc));
         }
      }
   ;

/**
 * Matches any ampersand prefaced symbolic name including all recognized or
 * unrecognized preprocessor directives. Such a statement starts with an
 * <code>'&'</code> followed by letters and possibly the hyphen character.
 * <p>
 * The resulting token's text will be compared with the literals table by
 * the <code>testLiteralsTable</code> method of this class. The matching is
 * done case-insensitively and a subset of the symbols can be matched with
 * an abbreviated form. The token type replacement occurs as follows:
 * <p>
 * <pre>
 * Token Type   Matched Text        Minimum Abbreviation Chars
 * -----------  ------------------  --------------------------
 * AGLOBAL      &global-define      5
 * ASCOPED      &scoped-define      5
 * AMESSAGE     &message            n/a
 * AUNDEFINE    &undefine           6
 * AIF          &if                 n/a
 * ATHEN        &then               n/a
 * AELSEIF      &elseif             n/a
 * AELSE        &else               n/a
 * AENDIF       &endif              n/a
 * ASUSPEND     &analyze-suspend    n/a
 * ARESUME      &analyze-resume     n/a
 * </pre> 
 * <p>
 * This is a top level lexer rule which means that in the case where the
 * statement is unrecognized, there will be an associated <code>ASTMT</code>
 * token.
 * <p>
 * If the leading ampersand is not followed by a recognized statement, then
 * the ampersand and any following alphabetic characters will be returned as a 
 * <code>CODE</code> token.  It is also possible to return the ampersand by
 * itself as a <code>CODE</code> token, if there are no following alphabetic
 * characters. Both of these combinations are possible in certain 4GL parsing
 * cases like shell commands the ampersand character is perfectly valid. Another
 * example is that ampersands can appear inside a variable name like my-p&l-var.
 * The 4GL is just insane.
 *
 * @throws   antlr.RecognitionException
 * @throws   antlr.CharStreamException
 * @throws   antlr.TokenStreamException
 */
ASTMT
   :
      '&'
      (
           // this lookahead will be too broad since it can match any
           // kind of alphabetic text, but when it matches anything that
           // isn't a known preproc statement, it will just be set to
           // code anyway
           { env.setInStatement(true); }
           (
                'a'..'z'
              | '-'
           )+
           
         | // nothing, this is really "code"
      )
      // this will be overridden by the literals testing when necessary
      { $setType(CODE); }
   ;

/**
 * Matches a preprocessor input that has no other interpretation.
 * Combines into one token as many characters as possible without
 * breaking other rules. Breaks at &, /, ", ', * and whitespace to avoid
 * lexical non-determinism with other top-level rules.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>CODE</code> token.
 */
CODE
   {
      boolean bogusStmt = false;
      boolean isComment = false;
      
      if (LA(1) == '/')
      {
         env.setDeferMarkers(true);
         bogusStmt = true;
      }
   }
   :
      (
          COMMENT  { $setType(COMMENT); isComment = true; }
          
        | { LA(2) != '*' }?
          '/'
          {
             if (bogusStmt)
             {
                env.setDeferMarkers(false);
                env.clearDeferred();
             }
          }
          
        | ~(' ' | '\t' | '\n' | '&' | '/' | '\"' | '\'')
      )
      (
           { !isComment }?
           (
              options { greedy = true; }
              :
              ~(' ' | '\t' | '\n' | '&' | '/' | '\"' | '\'' | '*')
           )*
           
         | // empty alternative for comment case
      )
   ;

/**
 * Matches Progress language comments, possibly nested. Brackets
 * the comments with a pair
 * of calls to <code>{@link Environment#setInComment(boolean)}</code>
 * so the preprocessor always knows the context.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>COMMENT</code> token.
 */
protected
COMMENT
   {
      commentNesting++;
   }
   :
      (
         COMM_OPEN
         {
            if (commentNesting == 1)
            {
               env.setDeferMarkers(false);
               env.clearDeferred();
               env.setInComment(true);
            }
         }
         (
            options
            {
               greedy = false;
               generateAmbigWarnings = false;
            }
            :
              COMMENT
              
            | '\n' { newline(); }
            
            | .
         )*
         COMM_CLOSE
      )
      {
         commentNesting--;
         
         // reset inComment flag at outermost nesting level only
         if (commentNesting == 0)
         {
            env.setInComment(false);
         }
      }
   ;

/**
 * Matches the opening sequence for Progress 4GL comments.
 */
protected
COMM_OPEN
   :
      '/' '*'
   ;

/**
 * Matches the closing sequence for Progress 4GL comments.
 */
protected
COMM_CLOSE
   :
      '*' '/'
   ;

/**
 * Matches an opening single quote, arbitrary contents and an ending single
 * quote.  Two continguous single quote characters and an escape prefixed
 * single quote character are accepted as contents (they do not terminate
 * the string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.  Note that all such newlines
 * are maintained in the output string because these are the escaped chars
 * that have been left behind by the preprocessor.  If such characters are not
 * escaped, then the preprocessor removes such chars (carriage returns and
 * line feeds).  This is how Progress 4GL handles these characters.  This
 * means that in a raw source file that has not been preprocessed, a string
 * literal can be split across any number of lines and the Progress 
 * preprocessor will put the string back together, ignoring the carriage
 * returns and newlines.
 * <p>
 * Tilde and the backslash can BOTH be Progress escape characters and as such,
 * they need special attention. Backslash is only honored if UNIX escapes
 * mode is set on. In particular, this rule must separately match 3 or 6
 * constructs as part of a string depending on which escape sequences are
 * honored.
 * <p>
 * <pre>
 *    ~'
 *    \'
 *    ~~
 *    ~\
 *    \\
 *    \~
 *    ~
 *    \
 * </pre>
 * <p>
 * The first 2 are a way of embedding a quote in a string.  The next 4 are
 * important because if this rule were to encounter '~~', '\\', '~\' or '\~'
 * (which are valid strings) the rule would consume the first escape and then
 * encounter an escaped quote which would not terminate the string properly,
 * leading to a non-ending string.  So this rule matches on any escaped escape
 * char to eliminate this situation.  Finally, a single ~ or \ that is not
 * followed by a &quot; or a duplicate escape char is matched as a single
 * character.  This is required (one may think that the closure rule
 * should handle this case) because when ANTLR sees a ~ or \ in the leftmost
 * position of the alternatives, it DOES NOT include it in the list of
 * 'everything that is not a closing quote'.  Likewise \n is not included.
 * Of course, these tilde constructions must be placed in a specific order
 * and if so, the ambiguity warnings that ANTLR reports can be disabled.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped single quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings.
 */
protected
ASTRING
   :
      {
         boolean honor = env.getOpt().isUnixEscapes();
      }
      (
         '\''                              // opening '
              (
                 options { generateAmbigWarnings = false; }
                 :
                   "~\'"                   // allow ~'
                 | "~~"                    // allow ~~ (consume as one)
                 | "~\\"                   // allow ~\ (consume as one)
                 | '~'                     // allow ~ by itself
                 | { honor }?
                   "\\\'"                  // allow \'
                 | { honor }?
                   "\\\\"                  // allow \\ (consume as one)
                 | { honor }?
                   "\\~"                   // allow \~ (consume as one)
                 | '\\'                    // allow \ by itself
                 | "\'\'"                  // allow ''
                 
                   // in broken mode, we are parsing a global or scoped define
                   // and the parsing must honor the behavior of parsing a
                   // string but it must also allow that string to abruptly
                   // end at a newline without any closing quote character (the
                   // closing quote is optional in that case); although it seems
                   // that we could avoid this rule completely in the define
                   // case, it turns out that we cannot because the string
                   // parsing is not only complex but it also changes the
                   // behavior of the preprocessor in ways that are required;
                   // in non-broken mode, we just match the newline and continue
                   // the string as normal
                 | { !brokenStringMatching }? 
                   '\n'     { newline(); } // otherwise allow NL                 
                 | ~'\''                   // anything else except closing '
              )*
         (
              { !brokenStringMatching }?
              '\''                         // closing '
            | { brokenStringMatching }?
              ('\'')?                      // optional closing '
         )
      )
   ;

/**
 * Matches an opening double quote, arbitrary contents and an ending double
 * quote.  Two continguous double quote characters and an escape prefixed
 * double quote character are accepted as contents (they do not terminate
 * the string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.  Note that all such newlines
 * are maintained in the output string because these are the escaped chars
 * that have been left behind by the preprocessor.  If such characters are not
 * escaped, then the preprocessor removes such chars (carriage returns and
 * line feeds).  This is how Progress 4GL handles these characters.  This
 * means that in a raw source file that has not been preprocessed, a string
 * literal can be split across any number of lines and the Progress 
 * preprocessor will put the string back together, ignoring the carriage
 * returns and newlines.
 * <p>
 * Tilde and the backslash can BOTH be Progress escape characters and as such,
 * they need special attention. Backslash is only honored if UNIX escapes
 * mode is set on. In particular, this rule must separately match 3 or 6
 * constructs as part of a string depending on which escape sequences are
 * honored.
 * <p>
 * <pre>
 *    ~&quot;
 *    \&quot;
 *    ~~
 *    ~\
 *    \\
 *    \~
 *    ~
 *    \
 * </pre>
 * <p>
 * The first 2 are a way of embedding a quote in a string.  The next 4 are
 * important because if this rule were to encounter '~~', '\\', '~\' or '\~'
 * (which are valid strings) the rule would consume the first escape and then
 * encounter an escaped quote which would not terminate the string properly,
 * leading to a non-ending string.  So this rule matches on any escaped escape
 * char to eliminate this situation.  Finally, a double ~ or \ that is not
 * followed by a &quot; or a duplicate escape char is matched as a double
 * character.  This is required (one may think that the closure rule
 * should handle this case) because when ANTLR sees a ~ or \ in the leftmost
 * position of the alternatives, it DOES NOT include it in the list of
 * 'everything that is not a closing quote'.  Likewise \n is not included.
 * Of course, these tilde constructions must be placed in a specific order
 * and if so, the ambiguity warnings that ANTLR reports can be disabled.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped double quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings.
 */
protected
QSTRING
   :
      {
         boolean honor = env.getOpt().isUnixEscapes();
      }
      (
         '\"'                              // opening "
              (
                 options { generateAmbigWarnings = false; }
                 :
                   "~\""                   // allow ~"
                 | "~~"                    // allow ~~ (consume as one)
                 | "~\\"                   // allow ~\ (consume as one)
                 | '~'                     // allow ~ by itself
                 | { honor }?
                   "\\\""                  // allow \"
                 | { honor }?
                   "\\\\"                  // allow \\ (consume as one)
                 | { honor }?
                   "\\~"                   // allow \~ (consume as one)
                 | '\\'                    // allow \ by itself
                 | "\"\""                  // allow ""
                 
                   // in broken mode, we are parsing a global or scoped define
                   // and the parsing must honor the behavior of parsing a
                   // string but it must also allow that string to abruptly
                   // end at a newline without any closing quote character (the
                   // closing quote is optional in that case); although it seems
                   // that we could avoid this rule completely in the define
                   // case, it turns out that we cannot because the string
                   // parsing is not only complex but it also changes the
                   // behavior of the preprocessor in ways that are required;
                   // in non-broken mode, we just match the newline and continue
                   // the string as normal
                 | { !brokenStringMatching }? 
                   '\n'     { newline(); } // otherwise allow NL
                 | ~'\"'                   // anything else except closing "
              )*
         (
              { !brokenStringMatching }?
              '\"'                         // closing "
            | { brokenStringMatching }?
              ('\"')?                      // optional closing "
         )
      )
   ;
