/*
** Module   : Variable.java
** Abstract : contains the data needed to define a Progress variable
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050210   @19752 First version, simple bean-like approach.
** 002 GES 20050330   @20563 Added support for a large number of options
**                           and helper methods to copy options and write
**                           them as annotations.
** 003 GES 20050401   @20597 Added support for a temporary index to allow
**                           the linking of a variable reference to its
**                           original definition at parse time. This is
**                           needed because project-unique IDs can't be
**                           easily generated at parse time, but are only
**                           added after the entire tree is complete.
** 004 GES 20050411   @20674 Massive rewrite to use ASTs to store variable
**                           options and to exchange ASTs with the schema
**                           dictionary and fully handle field and
**                           variable options.  Most of the simple getter/
**                           setter methods are removed since they are
**                           not generally needed.
** 005 GES 20050414   @20724 Cleaned up initializer type conversions and
**                           added a special booleanManhandle() method 
**                           to post-process a boolean initializer that
**                           is a user-defined string and thus cannot be
**                           properly lexed or parsed "in-line".
** 006 GES 20050501   @20956 Added annotations of the name and token type
**                           for the original variable definitions which
**                           eases processing at conversion time since
**                           the children don't have to be walked to get
**                           critical data.
** 007 GES 20070329   @32654 Fixed problem with some cases where extent 
**                           options not needed to be converted into a
**                           non-extent form before the options were
**                           returned.
** 008 GES 20070629   @34347 Minor interface change.
** 009 GES 20070919   @35140 Added support for the fully qualified name
**                           of the class represented by a VAR_CLASS.
** 010 GES 20080702   @39057 Number initializers can be strings.
** 011 ECF 20080704   @39147 Minor memory optimization. Use Boolean
**                           constants instead of instantiating new
**                           Boolean objects.
** 012 GES 20090424   @41958 Import change.
** 013 GES 20090429   @42068 Match package and class name changes.
** 014 GES 20110826          Avoid NPE when there is an unspecified extent.
**                           Added annotations for 10.2x features.
** 015 GES 20111202          Handle boolean initializers for numeric types.
*/

package com.goldencode.p2j.uast;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.util.*;

/**
 * A simple wrapper class to contain all related data about a variable. It
 * can also be used to store data related to more simple widgets and frames.
 * All options are stored as an AST and helper methods are provided to set,
 * get, merge and annotate using these option ASTs (which come directly
 * from the source trees or from the schema dictionary).
 * <p>
 * <b>One common option is unusual:</b> 
 * <p>
 * <ul>
 *     <li> VIEW-AS phrase (this is a complex subtree, not just a simple
 *          value so it is stored in the AST BUT is cannot be stored into
 *          an annotation at this time)
 * </ul>
 * <p>
 * All options that can be specified in a <code>DEFINE VARIABLE</code>
 * statement are supported except:
 * <p>
 * <ul>
 *     <li> BGCOLOR
 *     <li> CONTEXT-HELP-ID
 *     <li> DCOLOR
 *     <li> DROP-TARGET
 *     <li> FONT
 *     <li> FGCOLOR
 *     <li> MOUSE-POINTER
 *     <li> PFCOLOR
 * <p>
 * All options that can be specified in <code>DEFINE PARAMETER</code> are
 * supported (this is simply the parameter type itself).
 * <p>
 * All options that can be specified in a <code>format phrase</code> are
 * supported except:
 * <p>
 * <ul>
 *     <li> AUTO-RETURN 
 *     <li> BGCOLOR
 *     <li> BLANK
 *     <li> COLON
 *     <li> CONTEXT-HELP-ID
 *     <li> DCOLOR
 *     <li> DEBLANK
 *     <li> DISABLE-AUTO-ZAP
 *     <li> DROP-TARGET
 *     <li> FONT
 *     <li> FGCOLOR
 *     <li> MOUSE-POINTER
 *     <li> NO-TAB-STOP
 *     <li> PFCOLOR
 *     <li> TO
 *     <li> VALIDATE
 * </ul> 
 * <p>
 * All options that can be specified in a <code>message</code> statement are
 * supported except <code>AUTO-RETURN</code>.  
 *
 * @author    GES
 */
public class Variable
implements ProgressParserTokenTypes,
           TokenDataWrapper
{
   /** The full text name of the variable. */
   private String name = null;
   
   /**
    * Parser defined integer constant matching the token type of the
    * variable.
    */
   private int tokenType = -1;
   
   /**
    * Temporary index to link variable references with the original AST that
    * is the definition.  This should be set by the creator of the instance
    * before adding this to the dictionary.
    */
   private int tempIdx = -1;
   
   /** Stores the AST that roots all options associated with the variable. */
   private Aast def = null;
   
   /** Fully qualified class name of the object instance represented. */
   private String cls = null;
   
   /**
    * Constructs an instance with the required data to be contained.
    *
    * @param    name
    *           The full name of the variable.
    * @param    tokenType
    *           The token type (as defined by the parser) of the variable.
    * @param    cls
    *           Fully qualified class name of the object instance represented
    *           or <code>null</code> if this variable does not represent an
    *           object instance.
    */
   public Variable(String name, int tokenType, String cls)
   throws IllegalArgumentException
   {
      if (name == null || tokenType == -1)
         throw new IllegalArgumentException("Invalid name or type.");
      
      this.name      = name;
      this.tokenType = tokenType;
      this.cls       = cls;
   }
   
   /**
    * Make a copy of the given options, possibly making modifications on the
    * copy.  Specifically, if <code>extent</code> is greater than 0 and the
    * source field is an extent field, the copy will represent a single
    * element of the original extent field.  The effect of this is to remove
    * the KW_EXTENT child node, and to use only the initializer (if any)
    * which corresponds with the element at index <code>extent</code>.
    *
    * @param   source
    *          Options AST to be copied.
    * @param   extent
    *          One-based index of the element (for an extent field) after
    *          which the new field entry is to be modelled. This is only
    *          relevant for the initializer, so that we copy the correct
    *          initial value in the case of an initializer list for an extent
    *          field.  Will be <code>0</code> if the "like" field is not an
    *          extent field, or if the entire extent field is the target of
    *          the like phrase, rather than only a specific element.
    *
    * @return  Copy of the AST corresponding with the given name, with the
    *          possible modifications described above.
    */
   public static Aast copyOptions(Aast source, int extent)
   {
      Aast copy = source.duplicateFresh();
      
      // Handle special case where like target is a single element of an
      // extent field.
      if (extent > 0)
      {
         // we are changing some values, make sure no conflicting annotations
         // exist
         copy.clearAnnotations();
         
         // Remove KW_EXTENT child node.
         Aast kwExtent = copy.getImmediateChild(KW_EXTENT, null);
         if (kwExtent != null)
         {
            kwExtent.remove();
         }
         
         // Normalize index.
         extent--;
         
         // Remove all KW_INIT children which are not the target we need
         // to retain.
         Aast kwInit = copy.getImmediateChild(KW_INIT, null);
         if (kwInit != null && kwInit.getNumImmediateChildren() > 1)
         {
            int size = kwInit.getNumImmediateChildren();
            
            // Due to a bug in Progress 9.1c (and possibly other versions),
            // a temp-table field or var like some specific element of any
            // extent field/var always picks up the initial value of the
            // FIRST element in the extent, rather than the correct one.  We
            // emulate this behavior by always setting extent to 0.  The more
            // correct approach is commented out below, in case we need to do
            // it right for other versions.
            Aast initTgt = kwInit.getChildAt(0);
            /*
            ---- BEGIN BETTER APPROACH ----
            Aast initTgt = kwInit.getChildAt(extent);
            if (initTgt == null)
            {
               // We have more than one initializer value, but not enough
               // to match the index we were given, so use the last
               // available.
               initTgt = kwInit.getChildAt(size - 1);
            }
            ----- END BETTER APPROACH -----
            */
            
            // Remove all initializers from the KW_INIT node except the
            // intended target.
            Aast child = kwInit.getChildAt(0);
            
            while (child != null)
            {
               Aast next = (Aast) child.getNextSibling();
               
               if (child != initTgt)
               {
                  child.remove();
               }
               
               child = next; 
            }
         }
      }
      
      return copy;
   }
   
   /**
    * Returns the full name of the variable.
    *
    * @return   The full name of the variable.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Sets the full name of the variable.
    *
    * @param    name
    *           The full name of the variable.
    */
   public void setName(String name)
   {
      this.name = name;
   }
   
   /**
    * Returns the parser-defined token type of the variable.
    *
    * @return   The token type of the variable.
    */
   public int getTokenType()
   {
      return tokenType;
   }   
   
   /**
    * Sets the parser-defined token type of the variable.
    *
    * @param    tokenType
    *           The token type of the variable.
    */
   public void setTokenType(int tokenType)
   {
      this.tokenType = tokenType;
   }   
   
   /**
    * Returns the temporary index that identifies the original AST node that
    * defined the variable.
    *
    * @return   The temporary index.
    */
   public int getTempIndex()
   {
      return tempIdx;
   }   
   
   /**
    * Sets the temporary index that identifies the original AST node that
    * defined the variable.
    *
    * @param    tempIdx
    *           The temporary index.
    */
   public void setTempIndex(int tempIdx)
   {
      this.tempIdx = tempIdx;
   }   
      
   /**
    * Obtains the fully qualified name of the class of the object instance
    * represented by this variable.
    *
    * @return   The fully qualified name of the class of the object instance
    *           represented by this variable or <code>null</code> if this
    *           variable does not represent an object instance.
    */
   public String getClassName()
   {
      return cls;
   }
   
   /**
    * Sets the fully qualified name of the class of the object instance
    * represented by this variable.
    *
    * @param    cls
    *           The fully qualified name of the class of the object instance
    *           represented by this variable or <code>null</code> if this
    *           variable does not represent an object instance.
    */
   public void setClassName(String cls)
   {
      this.cls = cls;
   }
   
   /**
    * Copies all options based on another object of this same type.  The name,
    * token type and temporary index are NOT overridden. 
    *
    * @param    like
    *           The instance from which to copy the options.
    */
   public void copyDefaultOptions(Aast like)
   {
      if (like != null)
      {
         def = like.duplicate(null);
      }
   }
   
   /**
    * Sets all options based on the children of the passed-in AST, overriding
    * any options previously copied based on a <code>LIKE</code> clause.  The
    * name, token type and temporary index are NOT overridden.  <b>This merge 
    * process may create new AST nodes where none existed in the original
    * source tree!</b>
    * <p>
    * The following options are handled:
    * <p>
    * <ul>
    *    <li> FORMAT
    *    <li> INITIAL
    *    <li> LABEL
    *    <li> COLUMN-LABEL
    *    <li> DECIMALS
    *    <li> EXTENT
    *    <li> CASE-SENSITIVE
    *    <li> NOT CASE-SENSITIVE
    *    <li> VIEW-AS
    * </ul>
    *
    * @param    ast
    *           The AST whose children define the options to be set.
    */
   public void setOptions(Aast ast)
   {  
      // are there any previous options to consider?
      if (def == null)
      {
         // simple case: no previous like clause
         def = ast;
         return;
      }
      
      // merge old options into AST
      overrideIfNotPresent(def, ast, KW_FORMAT);
      overrideIfNotPresent(def, ast, KW_INIT);
      overrideIfNotPresent(def, ast, KW_LABEL);
      overrideIfNotPresent(def, ast, KW_COL_LAB);
      overrideIfNotPresent(def, ast, KW_DECIMALS);
      overrideIfNotPresent(def, ast, KW_EXTENT);
      overrideIfNotPresent(def, ast, KW_VIEW_AS);
      
      // special processing for case-sensitive (only copy if neither option
      // exists in the target AST)
      if (ast.getImmediateChild(KW_NOT, null)      == null &&
          ast.getImmediateChild(KW_CASE_SEN, null) == null)
      {
         // try copying the NOT CASE-SENSITIVE
         if (overrideIfNotPresent(def, ast, KW_NOT) == false)
         {
            // try copying the CASE-SENSITIVE
            overrideIfNotPresent(def, ast, KW_CASE_SEN);
         }      
      }      
      
      // save off the merged AST
      def = ast;
   }   
   
   /**
    * Gets the AST that defines all options for this variable. These options
    * may have been copied based on a <code>LIKE</code> clause or set via a
    * source code definition.  The name, token type and temporary index are 
    * NOT accessible via the returned AST. 
    * <p>
    * The returned options may have been modified. Specifically, if 
    * <code>extent</code> is greater than 0 and the source variable is an
    * extent variable, the result will represent a copy of a single element
    * of the original extent variable. The effect of this is to remove the
    * KW_EXTENT child node, and to use only the initializer (if any) which
    * corresponds with the element at index <code>extent</code>.
    *
    * @param   extent
    *          One-based index of the element (for an extent field) after
    *          which the new field entry is to be modelled. This is only
    *          relevant for the initializer, so that we copy the correct
    *          initial value in the case of an initializer list for an extent
    *          field.  Will be <code>0</code> if the "like" field is not an
    *          extent field, or if the entire extent field is the target of
    *          the like phrase, rather than only a specific element.
    *
    * @return   The AST whose children define the options for this variable.
    */
   public Aast getOptions(int extent)
   {
      int  len      = 0;
      Aast possible = def.getImmediateChild(KW_EXTENT, null);
      
      if (possible != null)
         len = Integer.parseInt(possible.getFirstChild().getText());
      
      // WARNING: should we only return a copy here? It doesn't seem to
      //          cause any trouble, but it seems dangerous. We could avoid
      //          all of this work and just let copyOptions() do the job...
      if (extent == 0 || possible == null || len == 0)
         return def;
      
      return copyOptions(def, extent);
   }   
   
   /**
    * Stores all non-standard options as annotations to the AST passed as
    * a parameter if this is the original definition, otherwise it stores
    * a subset of values, most important of which is the cross-reference
    * index to allow the original definition to be easily identified and
    * accessed from a reference.  If an option has the default value, it is
    * not saved as an annotation.
    *
    * @param    ast
    *           The AST to annotate.    
    * @param    original
    *           <code>true</code> if this is to annotate the AST of the
    *           original definition, <code>false</code> if the AST is only
    *           a reference.
    * @param    sym
    *           The current symbol resolver.
    */
   public void annotateOptions(Aast ast, boolean original, SymbolResolver sym)
   {
      if (ast == null)
      {
         // nothing to do
         return;
      }
      
      if (tempIdx != -1)
         ast.putAnnotation("tempidx", new Long(tempIdx));      
      
      if (tokenType == VAR_CLASS || tokenType == FIELD_CLASS)
      {
         sym.annotateClassRef(cls, ast);
      }
      
      // the following annotations are only stored in the original defining
      // AST rather than in references
      if (!original || def == null)
      {
         // nothing else to do
         return;
      }
      
      // storing the name and type in annotations makes variable definitions
      // easier to generate at conversion time
      if (name != null)
         ast.putAnnotation("name", name);      
      
      if (tokenType != -1)
         ast.putAnnotation("type", new Long(tokenType));      
      
      
      Aast child        = (Aast) def.getFirstChild();
      Aast grandchild;
      
      while (child != null)
      {
         int ctype = child.getType();
         
         switch (ctype)
         {
            case KW_SHARED:
               ast.putAnnotation("shared", Boolean.TRUE );
               
               boolean glob = def.getImmediateChild(KW_GLOBAL, null) != null;
               boolean newed  = def.getImmediateChild(KW_NEW, null) != null;
               
               ast.putAnnotation("global",
                                 glob ? Boolean.TRUE : Boolean.FALSE);
               ast.putAnnotation("newed",
                                 newed ? Boolean.TRUE : Boolean.FALSE);
               
               break;
            case KW_INPUT:
            case KW_IN_OUT:
            case KW_OUTPUT:
               ast.putAnnotation("parmtype", new Long(ctype));
               break;
            case KW_PUBLIC:
            case KW_PROTECTD:
            case KW_PRIVATE:
               ast.putAnnotation("access-mode", new Long(ctype));
               break;
            case KW_STATIC:
               ast.putAnnotation("static", new Boolean(true));
               break;
            case KW_ABSTRACT:
               ast.putAnnotation("abstract", new Boolean(true));
               break;
            case KW_OVERRIDE:
               ast.putAnnotation("override", new Boolean(true));
               break;
            case KW_FORMAT:
               grandchild = (Aast) child.getFirstChild();
               ast.putAnnotation("format",
                                 grandchild.getFirstChild().getText());
               break;
            case KW_HELP:
               ast.putAnnotation("help", child.getFirstChild().getText());
               break;
            case KW_INIT:
               String name     = "initial"; 
               int  j          = 0;
               grandchild = (Aast) child.getFirstChild();
               
               while (grandchild != null)
               {
                  int     gtype   = grandchild.getType();
                  String  gtxt    = grandchild.getText();
                  boolean unknown = (gtype == UNKNOWN_VAL);
                  
                  switch (tokenType)
                  {
                     case VAR_DEC:
                        // this will handle NUM_LITERAL and DEC_LITERAL which
                        // are both valid initializers in Progress
                        Double d = null;
                        
                        if (!unknown)
                        {
                           if (gtype == NUM_LITERAL || gtype == DEC_LITERAL)
                           {
                                 d = new Double(gtxt);
                           }
                           else if (gtype == STRING)
                           {
                              gtxt = character.progressToJavaString(gtxt);
                              
                              if (gtxt.length() == 0)
                              {
                                 d = new Double(0);
                              }
                              else
                              {
                                 d = new Double(gtxt);
                              }
                           }
                           else if (gtype == BOOL_TRUE || gtype == BOOL_FALSE)
                           {
                              // Progress silently accepts this initializer, we
                              // assume it is the same as 0 (no init), but this
                              // is untested at this time.
                              d = new Double(0);
                           }
                        }
                        
                        ast.putAnnotation(name, d, j);
                        break;
                     case VAR_INT:
                        Long l = null;
                     
                        if (!unknown)
                        {
                           if (gtype == DEC_LITERAL)                     
                           {
                              // Progress allows decimal initializers for
                              // integers, in this case we must round to the
                              // nearest integer...
                              Double dd = new Double(gtxt);
                              l = new Long(Math.round(dd.doubleValue()));
                           }
                           else if (gtype == NUM_LITERAL)
                           {
                              l = new Long(gtxt);
                           }
                           else if (gtype == STRING)
                           {
                              gtxt = character.progressToJavaString(gtxt);
                              
                              if (gtxt.length() == 0)
                              {
                                 l = new Long(0);
                              }
                              else
                              {
                                 l = new Long(gtxt);
                              }
                           }
                           else if (gtype == BOOL_TRUE ||
                                    gtype == BOOL_FALSE)
                           {
                              // Progress silently accepts this initializer,
                              // we assume it is the same as 0 (no init), but
                              // this is untested at this time.
                              l = new Long(0);
                           }
                        }
                        
                        ast.putAnnotation(name, l, j);
                        break;
                     case VAR_LOGICAL:
                        // fixup stupid Progress "feature"
                        if (gtype == STRING)
                        {
                           try
                           {
                              booleanManhandle(def, grandchild);
                           }
                           
                           catch (AstException ae)
                           {
                              // TODO: log this
                              ae.printStackTrace();
                           }
                        }

                        Boolean b = null; 
                        
                        if (!unknown)
                        {
                           b = (grandchild.getType() == BOOL_TRUE ?
                                Boolean.TRUE : Boolean.FALSE);
                        }
                        
                        ast.putAnnotation(name, b, j);
                        break;
                     case VAR_CHAR:
                     case VAR_DATE:
                        String s = unknown ? 
                                   null    :
                                   new String(grandchild.getText());
                        
                        ast.putAnnotation(name, s, j);
                        break;
                  }
                  j++;
                  grandchild = (Aast) grandchild.getNextSibling();
               }
               break;
            case KW_LABEL:
               int  i          = 0;
               grandchild = (Aast) child.getFirstChild();
               
               while (grandchild != null)
               {
                  ast.putAnnotation("label", grandchild.getText(), i);
                  i++;
                  grandchild = (Aast) grandchild.getNextSibling();
               }
               break;
            case KW_COL_LAB:
               ast.putAnnotation("columnlabel",
                                 child.getFirstChild().getText());
               break;
            case KW_DECIMALS:
               int prec = Integer.parseInt(child.getFirstChild().getText());
               if (tokenType == VAR_DEC && prec != 10)
               {
                  ast.putAnnotation("decimals", new Long(prec));
               }
               break;
            case KW_EXTENT:
               Aast extChild = child.getChildAt(0);
               
               int ext = -1;  // an unspecified extent is -1 (TODO: is that OK?)
               if (extChild != null)
                  ext = Integer.parseInt(extChild.getText());
               
               ast.putAnnotation("extent", new Long(ext));
               break;
            case KW_NOT:
               if (child.getFirstChild().getType() == KW_CASE_SEN)
               {
                  ast.putAnnotation("casesens", Boolean.FALSE);
               }
               break;
            case KW_CASE_SEN:
               ast.putAnnotation("casesens", Boolean.TRUE);
               break;
            case KW_NO_UNDO:
               ast.putAnnotation("undoable", Boolean.FALSE);
               break;
         }
         
         child = (Aast) child.getNextSibling();
      }      
   }
   
   /**
    * This method provides a fixup for a ridiculous Progress feature of
    * allowing arbitrary user-defined initializer strings (that are an exact
    * match or an unambiguous abbreviated match) to a format string of the
    * form 'XXX/YYY' where x, X, xx, XX, xxx, XXX all should resolve to 
    * <code>BOOL_TRUE</code> and y, Y, yy, YY, yyy, YYY all should resolve to
    * <code>BOOL_FALSE</code>.  This processing can't be done during the
    * initial parsing because the format string can appear <b>after</b> the
    * initializer!  
    * <p>
    * If no format string is available, then the default is made to be
    * 'yes/no' and 'true/false'.  Otherwise the specified format is used.
    * <p>
    * This method will rewrite input AST's token type to be the proper
    * boolean literal token type.
    *
    * @param    grandparent
    *           This is the anscestor 2 levels up from the node being
    *           processed (bogusString). It must be passed in because
    *           this processing occurs during parsing, before the parent
    *           linkages have been fixed up in the tree.  For this reason,
    *           it is not possible to use {@link Aast#getAncestor} or other
    *           more natural approaches to obtain the grandparent.
    * @param    bogusString
    *           The AST node which is a <code>STRING</code> but should be a
    *           <code>BOOL_TRUE</code> or <code>BOOL_FALSE</code>.
    *
    * @throws   AstException 
    *           When no match is found.
    */
   public void booleanManhandle(Aast grandparent, Aast bogusString)
   throws AstException
   {
      Aast   format      = grandparent.getImmediateChild(KW_FORMAT, null);
      String fmt         = null;
      
      ArrayList list = new ArrayList();
      list.add("yes/no");
      list.add("true/false");
      
      if (format != null)
      {
         fmt = format.getFirstChild().getFirstChild().getText().toLowerCase();
         list.add(StringHelper.stripEnclosing(fmt, '\"'));
      }
      
      String[] fmts   = (String[]) list.toArray(new String[0]);
      String   target = bogusString.getText().toLowerCase();
      
      target = StringHelper.stripEnclosing(target, '\"');
      
      for (int i = 0; i < fmts.length; i++)
      {
         String current = fmts[i];
         int    match   = current.indexOf(target);
         
         // check if there was a match
         if (match >= 0)
         {   
            if (match == 0)
            {
               // it is a match at the beginning of the string (everything
               // before the '/' is true)
               bogusString.setType(BOOL_TRUE);
            }
            else
            {
               bogusString.setType(BOOL_FALSE);
            }               
            
            // no more searching
            break;
         }
      }      
      
      int newtype = bogusString.getType();
      if (newtype != BOOL_TRUE && newtype != BOOL_FALSE) 
      {
         // this should never happen!
         throw new AstException("VAR_LOGICAL of unknown format (text " + 
                                 bogusString.getText() +
                                 ", format string = " + fmt + ") ");
      }
   }
   
   /**
    * Helper to copy a child subtree of a specific type from the source to
    * the target AST if the target AST does not already have a child of that
    * type.
    *
    * @param    source
    *           The AST to copy from.
    * @param    target
    *           The AST to copy to.
    * @param    type
    *           The token type of the child tree to copy if it exists in the
    *           source tree and does not already exist in the target tree.
    *
    * @return   <code>true</code> if a copy occurred.
    */
   private boolean overrideIfNotPresent(Aast source, Aast target, int type)
   {
      Aast child = target.getImmediateChild(type, null); 
      
      if (child == null)
      {
         child = source.getImmediateChild(type, null);
         
         if (child != null)
         {
            try
            {
               child.graftCopyTo(target, (int[]) null, false);
            }
            
            catch (AstException ae)
            {
               // unexpected failure
               return false;
            }
            
            return true;
         }
      }
      
      return false;
   }
}
