/*
** Module   : Statistic.java
** Abstract : counts or accumulates matches to a specific condition in some
**            list of files
**
** Copyright (c) 2005, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050225   @20123 Created initial version supporting 
**                           maintaining a a map of files and for each
**                           file a set of matches.  Counters are kept 
**                           for the total number of matches, the
**                           number of files with matches and all of the 
**                           backing data is made accessible.
** 002 GES 20050318   @20440 Added support to save off the condition
**                           during match processing and to provide an
**                           iterator to these strings for future
**                           reporting use.
** 003 ECF 20111206          Added ID for database support.
*/ 

package com.goldencode.p2j.util;

import java.util.*;

/**
 * Counts the number of matches of a specific named condition by file. 
 *
 * @author    GES
 * @version   1.0.0
 */
public class Statistic
{
   /** Next statistic ID (for database use) */
   private static int nextID = 0;
   
   /** Database ID for this statistic */
   private final int id;
   
   /** The verbose descriptive name of this statistic, used in reports. */
   private String descr = null;
   
   /** Counter that records the total number of matches across all files. */
   private int aggregate = 0;

   /** Counter that records the total number files with matches. */
   private int matchFiles = 0;

   /**
    * The list of files for each of which an associated set of matches is
    * stored.
    */
//   private Map files = new LinkedHashMap();
   private Map<String, Integer> files = new LinkedHashMap<String, Integer>();
   
   /**
    * Stores all conditions (expressions) that were matched for this
    * statistic.
    */
   private Set<String> conditions = new LinkedHashSet<String>();
   
   /**
    * Constructs an instance of a cross-file statistic with a specific
    * descriptive name.
    *
    * @param    descr
    *           Verbose description of this statistic.
    */
   public Statistic(String descr)
   {
      this.id = ++nextID;
      this.descr = descr;
   }
   
   /**
    * Accesses the database ID for this statistic
    * 
    * @return   The database ID for this statistic.
    */
   public int getID()
   {
      return id;
   }
   
   /**
    * Accesses the verbose description of this statistic.
    *
    * @return   The verbose description of this statistic.
    */
   public String getDescr()
   {
      return descr;
   }
   
   /**
    * Sets the verbose description of this statistic.
    *
    * @param    descr
    *           The verbose description of this statistic.
    */
   public void setDescr(String descr)
   {
      this.descr = descr;
   }
   
   /**
    * Add a match at a specific location in a specific file to the list.
    *
    * @param    filename
    *           File name of the file in which the match was found. This is
    *           used as a key and must be the same for all matches in this
    *           file.
    * @param    condition
    *           The expression that was evaluated to match this statistic.
    *           This may have been a positive match or a negative match.
    * @param    line
    *           Line number on which the match starts.
    * @param    column
    *           Column number on which the match starts.
    * @param    text
    *           User-defined text of the match.
    */
   public void match(String filename,
                     String condition,
                     int    line, 
                     int    column,
                     String text)
   {
//      Match match = new Match(line, column, text);
      
//      LinkedHashSet mset = (LinkedHashSet) files.get(filename);
      Integer matchCount = files.get(filename);
      
      // does the file already exist in our list?
//      if (mset == null)
//      {
         // a new file is being processed
//         mset = new LinkedHashSet();
//         files.put(filename, mset);         
//      }
      
      // we can have a file in our list without any matches... no text = no
      // match
      if (text != null)
      {
         // increment the total number of matches
         aggregate++;
         
         // if this is the first addition to a given file's set, increment
         // the number of files that have matches
//         if (mset.size() == 0)
         if (matchCount == null)
            matchFiles++;
         
         // add the match
//         mset.add(match);
         matchCount = (matchCount != null ? matchCount + 1 : 1);
         files.put(filename, matchCount);
         
         // remember the condition that triggered this match
         conditions.add(condition);
      }
   }
   
   /**
    * Accesses the total number of matches across all files.
    *
    * @return   The total number of matches across all files.
    */
   public int getAggregateCounter()
   {
      return aggregate;
   }
   
   /**
    * Accesses the number of files which have matches.  Some files in
    * the list may have no matches.
    *
    * @return   The number of files in the list which have matches.
    */
   public int getMatchFiles()
   {
      return matchFiles;
   }
   
   /**
    * Accesses the total number of files in the list.  Some files in the
    * list may have no matches.
    *
    * @return   The total number of files in the list.
    */
   public int getTotalNumberOfFiles()
   {
      return files.size();
   }
   
   /**
    * Accesses the iterator to the list of file-specific match sets.  This
    * allows direct access to the <code>Set</code> objects that each contain
    * all matches for each given file. 
    *
    * @return   The iterator to the list of match sets.
    */
   public Iterator<String> getFileIterator()
   {
      return files.keySet().iterator();
   }
   
   /**
    * Returns the <code>Set</code> of all known filenames.  Each filename
    * can be used as a key to obtain a list of matches (see
    * {@link #getMatchesFor}).
    *
    * @return   The set of all filenames.
    */
   /*
   public Set<String> getFilenames()
   {
      return files.keySet();
   }
   */
   
   /**
    * Returns an iterator to all known conditions that triggered a match,
    * each condition is a string that encodes the original expression 
    * found in the pattern engine rule that matched this statistic.
    *
    * @return   The set of all conditions (expressions).
    */
   public Iterator<String> getConditionsIterator()
   {
      return conditions.iterator();
   }
   
   /**
    * Returns the number of matches for a given file.
    *
    * @return   The number of matches for a given file or 0 if the file is
    *           not in the file list.
    */
   public int getCounterFor(String filename)
   {
      Integer matchCount = files.get(filename);
      
      // safety first
      if (matchCount == null)
      {
         return 0;
      }
      
      return matchCount;
      
//      LinkedHashSet mset = (LinkedHashSet) files.get(filename);
      
      // safety first
//      if (mset == null)
//         return 0;
      
//      return mset.size();
   }
   
   /**
    * Returns an iterator to the matches for a given file.
    *
    * @return   An iterator for the matches in a given file.
    */
   /*
   public Iterator getMatchesFor(String filename)
   {
      LinkedHashSet mset = (LinkedHashSet) files.get(filename);
      return mset.iterator();
   }
   */
}
